/*
 * Copyright 2005,2006 WSO2, Inc. http://www.wso2.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.solutions.identity.persistence;

import java.util.Date;

import org.wso2.solutions.identity.IdentityProviderException;
import org.wso2.solutions.identity.persistence.dao.ActionDAO;
import org.wso2.solutions.identity.persistence.dao.BaseDAO;
import org.wso2.solutions.identity.persistence.dao.ClaimDAO;
import org.wso2.solutions.identity.persistence.dao.ClaimValueDAO;
import org.wso2.solutions.identity.persistence.dao.DialectDAO;
import org.wso2.solutions.identity.persistence.dao.InfoCardDAO;
import org.wso2.solutions.identity.persistence.dao.IsssuedTokensDAO;
import org.wso2.solutions.identity.persistence.dao.PPIDValueDAO;
import org.wso2.solutions.identity.persistence.dao.ParameterDAO;
import org.wso2.solutions.identity.persistence.dao.RealmConfigurationDAO;
import org.wso2.solutions.identity.persistence.dao.RealmDAO;
import org.wso2.solutions.identity.persistence.dao.RegisteredInfoCardInfoDAO;
import org.wso2.solutions.identity.persistence.dao.RelyingPartyDAO;
import org.wso2.solutions.identity.persistence.dao.RevokedInfoCardDAO;
import org.wso2.solutions.identity.persistence.dao.UserPersonalRelyingPartyDAO;
import org.wso2.solutions.identity.persistence.dao.VerifierServiceDAO;
import org.wso2.solutions.identity.persistence.dataobject.AbstractDataObject;
import org.wso2.solutions.identity.persistence.dataobject.ActionDO;
import org.wso2.solutions.identity.persistence.dataobject.ClaimDO;
import org.wso2.solutions.identity.persistence.dataobject.ClaimValueDO;
import org.wso2.solutions.identity.persistence.dataobject.DialectDO;
import org.wso2.solutions.identity.persistence.dataobject.InfoCardDO;
import org.wso2.solutions.identity.persistence.dataobject.IssuedTokensDO;
import org.wso2.solutions.identity.persistence.dataobject.PPIDValueDO;
import org.wso2.solutions.identity.persistence.dataobject.ParameterDO;
import org.wso2.solutions.identity.persistence.dataobject.RealmConfigurationDO;
import org.wso2.solutions.identity.persistence.dataobject.RealmDO;
import org.wso2.solutions.identity.persistence.dataobject.RegisteredInfoCardInfoDO;
import org.wso2.solutions.identity.persistence.dataobject.RelyingPartyDO;
import org.wso2.solutions.identity.persistence.dataobject.RevokedInfoCardDO;
import org.wso2.solutions.identity.persistence.dataobject.UserPersonalRelyingPartyDO;
import org.wso2.solutions.identity.persistence.dataobject.VerifierServiceIdpDO;
import org.wso2.solutions.identity.persistence.dataobject.VerifierServiceRPDO;

/**
 * Handles all database persistence related operations of WSO2 Identity Solution.
 */
public class IPPersistenceManager {

    private static final String DEFAULT_CONFIG_FILE_NAME = "wso2identity-hibernate.cfg.xml";

    /**
     * Hibernate configuration instance to access the database.
     */
    HibernateConfig hbConfig = null;

    /**
     * Singleton instance of the persistence manager.
     */
    private static IPPersistenceManager manager;

    private IPPersistenceManager() throws IdentityProviderException {
        hbConfig = new HibernateConfig(DEFAULT_CONFIG_FILE_NAME);
    }

    private IPPersistenceManager(String configFile)
            throws IdentityProviderException {
        hbConfig = new HibernateConfig(configFile);
    }

    public static IPPersistenceManager getPersistanceManager()
            throws IdentityProviderException {
        if (manager == null) {
            manager = new IPPersistenceManager(DEFAULT_CONFIG_FILE_NAME);
        }
        return manager;
    }

    /**
     * Create a new entry with the given data object.
     * 
     * @param abstractDO
     *            Data object to be added
     * @return Identifier of the created entry.
     * @throws IdentityProviderException
     */
    public Long create(AbstractDataObject abstractDO)
            throws IdentityProviderException {
        try {
            BaseDAO base = new BaseDAO(hbConfig);
            return base.create(abstractDO);
        } catch (DuplicateEntityException e) {
            throw new IdentityProviderException("databaseAccessError", e);
        }
    }

    /**
     * Update an entry with the given data object.
     * 
     * @param abstractDO
     *            Data object to be updated
     */
    public void update(AbstractDataObject abstractDO) {
        BaseDAO base = new BaseDAO(hbConfig);
        base.update(abstractDO);
    }

    /**
     * Delete the entry of the given data object.
     * 
     * @param abstractDO
     *            Data object to be deleted.
     */
    public void delete(AbstractDataObject abstractDO) {
        BaseDAO base = new BaseDAO(hbConfig);
        base.delete(abstractDO);
    }

    /**
     * Extract a data object of the given type bearing the given id.
     * 
     * @param objClassName
     *            Type of data object
     * @param val
     *            Identifier of the record
     * @return If found, the requested data object
     */
    public AbstractDataObject getDataObject(String objClassName, Long val) {
        BaseDAO base = new BaseDAO(hbConfig);
        return base.getDataObject(objClassName, val);
    }

    /**
     * Provides all supported claims.
     * 
     * @return An array of
     *         <code>org.wso2.solutions.identity.persistence.dataobject.ClaimDO</code>
     *         instances.
     */
    public ClaimDO[] getAllSupportedClaims() {
        ClaimDAO dao = new ClaimDAO(hbConfig);
        return dao.getAllClaims();
    }

    /**
     * Provides all claims that are mapped to attributes in a user store.
     * 
     * @return An array of
     *         <code>org.wso2.solutions.identity.persistence.dataobject.ClaimDO</code>
     *         instances.
     */
    public ClaimDO[] getAllMappedClaims() {
        ClaimDAO dao = new ClaimDAO(hbConfig);
        return dao.getAllMappedClaims();
    }

    /**
     * Provides all claims that are mapped and enabled.
     * 
     * @return An array of
     *         <code>org.wso2.solutions.identity.persistence.dataobject.ClaimDO</code>
     *         instances.
     */
    public ClaimDO[] getAllMappedEnabledClaims() {
        ClaimDAO dao = new ClaimDAO(hbConfig);
        return dao.getAllMappedEnabledClaims();
    }

    /**
     * Provides all supported claim dialects.
     * 
     * @return An array of
     *         <code>org.wso2.solutions.identity.persistence.dataobject.DialectDO.java</code>
     *         instances.
     */
    public DialectDO[] getAllSupportedDialects() {
        DialectDAO dao = new DialectDAO(hbConfig);
        return dao.getAllDialects();
    }

    /**
     * Provides all claim values for a given user.
     * 
     * @param username
     *            The user identifier.
     * @return An array of
     *         <code>org.wso2.solutions.identity.persistence.dataobject.ClaimValueDO.java</code>
     *         instances.
     */
    public ClaimValueDO[] getClaimValuesForUser(String username) {
        ClaimValueDAO dao = new ClaimValueDAO(hbConfig);
        return dao.getClaimValuesForUser(username);
    }

    /**
     * Provides all PPID values in tokens issued for a given user.
     * 
     * @param userid
     *            The user identifier.
     * @return An array of
     *         <code>org.wso2.solutions.identity.persistence.dataobject.PPIDValueDO</code>
     *         instances.
     */
    public PPIDValueDO[] getPPIDValuesForUser(String userid) {
        PPIDValueDAO dao = new PPIDValueDAO(hbConfig);
        return dao.getPPIDValuesForUser(userid);
    }

    public InfoCardDO[] getCardIdsForUser(String username) {
        InfoCardDAO dao = new InfoCardDAO(hbConfig);
        return dao.getCardIdsForUser(username);
    }

    public int deleteCardsofUser(String username) {
        String stmt = "delete InfoCardDO where userId = :name";
        BaseDAO dao = new BaseDAO(hbConfig);
        return dao.executeSingleDMLStatement(stmt, "name", username);
    }

    public void getCardsIssuedBetween(Date startDate, Date endDate) {
        InfoCardDAO dao = new InfoCardDAO(hbConfig);
        dao.getCardsIssuedBetween(startDate, endDate);
    }

    public void backupExpiredCards() {
        InfoCardDAO dao = new InfoCardDAO(hbConfig);
        dao.backupExpiredCards();
    }

    public IssuedTokensDO[] getTokensOfUser(String username) {
        IsssuedTokensDAO dao = new IsssuedTokensDAO(hbConfig);
        return dao.getAllTokensOfUser(username);
    }

    public int deleteTokensOfUser(String username) {
        String stmt = "delete IssuedTokensDO where userId = :name";
        BaseDAO dao = new BaseDAO(hbConfig);
        return dao.executeSingleDMLStatement(stmt, "name", username);
    }

    public int deleteTokensForCard(String cardId) {
        String stmt = "delete IssuedTokensDO where cardId = :name";
        BaseDAO dao = new BaseDAO(hbConfig);
        return dao.executeSingleDMLStatement(stmt, "name", cardId);
    }

    public RelyingPartyDO[] getAllRelyingParties() {
        RelyingPartyDAO dao = new RelyingPartyDAO(hbConfig);
        return dao.getAllRelyingParties();
    }

    public RelyingPartyDO getRelyingParty(String url) {
        return (new RelyingPartyDAO(hbConfig)).getRelyingPartyDO(url);
    }

    public UserPersonalRelyingPartyDO getPersonalRelyingParty(String userId,
            String hostName) {
        return (new UserPersonalRelyingPartyDAO(hbConfig))
                .getPersonalRelyingParty(userId, hostName);
    }
    
    public ClaimDO getClaimByURI(String value) {
        return (new ClaimDAO(hbConfig)).getClaimByURI(value);
    }

    public InfoCardDO[] getAllInfoCards() {
        InfoCardDAO dao = new InfoCardDAO(hbConfig);
        return dao.getAllIssuedCards();
    }

    public PPIDValueDO[] getAllPPIDs() {
        PPIDValueDAO dao = new PPIDValueDAO(hbConfig);
        return dao.getAllPPIDs();
    }

    public InfoCardDO getInfoCard(String cardId) {
        InfoCardDAO dao = new InfoCardDAO(hbConfig);
        return dao.getInfoCard(cardId);
    }

    public Long createOrUpdateParameter(ParameterDO instance) {
        ParameterDAO dao = new ParameterDAO(hbConfig);
        return dao.createOrUpdateParameter(instance);
    }

    public ParameterDO getParameter(String name) {
        ParameterDAO dao = new ParameterDAO(hbConfig);
        return dao.getParameter(name);

    }

    public RealmConfigurationDO getEffectiveRealmConfiguration() {
        RealmConfigurationDAO dao = new RealmConfigurationDAO(hbConfig);
        return dao.getEffectiveRealmConfiguration();
    }

    public RealmConfigurationDO getRealmConfiguration(String name) {
        RealmConfigurationDAO dao = new RealmConfigurationDAO(hbConfig);
        return dao.getRealmConfiguration(name);
    }

    public RealmConfigurationDO[] getAllRealmConfigurations() {
        RealmConfigurationDAO dao = new RealmConfigurationDAO(hbConfig);
        return dao.getAllRealmConfigurations();
    }

    public RealmDO[] getAllRealms() {
        RealmDAO dao = new RealmDAO(hbConfig);
        return dao.getAllRealms();
    }

    public RealmDO getRealm(String className) {
        RealmDAO dao = new RealmDAO(hbConfig);
        return dao.getRealm(className);
    }

    public VerifierServiceIdpDO[] getAllVerifierServiceIdpDO() {
        VerifierServiceDAO dao = new VerifierServiceDAO(hbConfig);
        return dao.getAllVerifierServiceIdpDO();
    }

    public VerifierServiceRPDO[] getAllVerifierServiceRpDO() {
        VerifierServiceDAO dao = new VerifierServiceDAO(hbConfig);
        return dao.getAllVerifierServiceRpDO();
    }

    public void RevokeInfoCard(InfoCardDO infoCard)
            throws IdentityProviderException {
        InfoCardDAO icDao = new InfoCardDAO(hbConfig);
        icDao.delete(infoCard);
        RevokedInfoCardDO icDo = new RevokedInfoCardDO();
        icDo.setCardId(infoCard.getCardId());
        icDo.setDateIssued(infoCard.getDateIssued());
        icDo.setDateExpires(infoCard.getDateExpires());
        icDo.setUserId(infoCard.getUserId());
        icDo.setDateRevoked(new Date());
        create(icDo);
    }

    public RevokedInfoCardDO[] getAllRevokedInfoCards() {
        RevokedInfoCardDAO dao = new RevokedInfoCardDAO(hbConfig);
        return dao.getAllRevokedInfoCards();
    }

    public UserPersonalRelyingPartyDO[] getAllPersonalRelyingParties(
            String userId) {
        UserPersonalRelyingPartyDAO dao = new UserPersonalRelyingPartyDAO(
                hbConfig);
        return dao.getAllPersonalRelyingParties(userId);
    }

    public DialectDO getDialect(String uri) {
        DialectDAO dao = new DialectDAO(hbConfig);
        return dao.getDialect(uri);
    }

    public ActionDO[] getAllActions() {
        ActionDAO dao = new ActionDAO(hbConfig);
        return dao.getAllActions();
    }

    public RegisteredInfoCardInfoDO[] getAllRegistedInfoCardInfoForUser(
            String userId) {
        RegisteredInfoCardInfoDAO dao = new RegisteredInfoCardInfoDAO(hbConfig);
        return dao.getAllInfoForUser(userId);
    }

    public RegisteredInfoCardInfoDO getRegisteredInfoCardInfo(String ppid) {
        RegisteredInfoCardInfoDAO dao = new RegisteredInfoCardInfoDAO(hbConfig);
        return dao.getInformationCardInfo(ppid);
    }

}
