/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_tables.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "mod_cspace.h"
#include "cspace_utils.h"
#include "session.h"
#include "cspace_tvs.h"
#include "process_request.h"

#define SESSION_ID_LEN 64

#define MYMIN(a, b) ((a)<(b)?(a):(b))

#define UNAUTHORIZED_REDIRECT(r, u) redirect((r), (u))

/* TODO:This function should ideally redirect the browser to the url specified
 * by url. This is only used to redirect the browser to the login page when an
 * HTTP_UNAUTHORIZED occurs.
 *
 * We can live with the plain HTTP_UNAUTHORIZED for the moment
 */ 
 static int redirect(request_rec *r, char* url)
{
   return HTTP_UNAUTHORIZED;
}


static int find_uri(const char *uri, apr_array_header_t *a)
{
    int i;
    void *tmp;
    tmp = a->elts;
    for (i = 0; i < a->nelts; i++) {
#ifdef CSPACE_DEBUG
        printf("comparing:%s (%d), %s (%d), min=%d\n", *(char **)tmp,
                strlen(*(char **)tmp), uri, strlen(uri),
                MYMIN(strlen(*(char **)tmp), strlen(uri)));
#endif
        if (strncmp(*(char **)tmp, uri,
                    MYMIN(strlen(*(char **)tmp), strlen(uri))) == 0)
            return SUCC;
        tmp = (char *)tmp + a->elt_size;
    }

    return FAIL;
}

/*This piece of code is used only for testing purposes*/

/*
static int cspace_access_check(request_rec *r) {
    printf("cspace_access_check:\n");

    cspace_dir_cfg *dir_cfg = (cspace_dir_cfg *)ap_get_module_config(
            r->per_dir_config, &cspace_module);
    
    cspace_svr_cfg *svr_cfg = (cspace_svr_cfg *)ap_get_module_config(
            r->server->module_config, &cspace_module);
    
    cfg_svr_printf("\tchk_svr_cfg", svr_cfg, "\n");

    cfg_dir_printf("\tchk_dir_cfg", dir_cfg, "\n");
    return DECLINED;
}
*/

/* FROM http_filter.c:
 *
 * Here we deal with getting the request message body from the client.
 * Whether or not the request contains a body is signaled by the presence
 * of a non-zero Content-Length or by a Transfer-Encoding: chunked.
 *
 * Note that this is more complicated than it was in Apache 1.1 and prior
 * versions, because chunked support means that the module does less.
 *
 * The proper procedure is this:
 *
 * 1. Call ap_setup_client_block() near the beginning of the request
 *    handler. This will set up all the necessary properties, and will
 *    return either OK, or an error code. If the latter, the module should
 *    return that error code. The second parameter selects the policy to
 *    apply if the request message indicates a body, and how a chunked
 *    transfer-coding should be interpreted. Choose one of
 *
 *    REQUEST_NO_BODY          Send 413 error if message has any body
 *    REQUEST_CHUNKED_ERROR    Send 411 error if body without Content-Length
 *    REQUEST_CHUNKED_DECHUNK  If chunked, remove the chunks for me.
 *    REQUEST_CHUNKED_PASS     If chunked, pass the chunk headers with body.
 *
 *    In order to use the last two options, the caller MUST provide a buffer
 *    large enough to hold a chunk-size line, including any extensions.
 *
 * 2. When you are ready to read a body (if any), call ap_should_client_block().
 *    This will tell the module whether or not to read input. If it is 0,
 *    the module should assume that there is no message body to read.
 *
 * 3. Finally, call ap_get_client_block in a loop. Pass it a buffer and its
 *    size. It will put data into the buffer (not necessarily a full buffer),
 *    and return the length of the input block. When it is done reading, it will
 *    return 0 if EOF, or -1 if there was an error.
 *    If an error occurs on input, we force an end to keepalive.
 *
 *    This step also sends a 100 Continue response to HTTP/1.1 clients if
 *    appropriate.
 */

static int cspace_read_request(request_rec *r, char **buf, int *retval)
{
    apr_array_header_t *a;
    apr_pool_t *p = r->pool;
    apr_pool_t *temp_pool;
    int res, ret;
    
#ifdef CSPACE_DEBUG
    printf("CSPACE_READ_REQUEST:\n");
#endif

    if (apr_pool_create(&temp_pool, p) != APR_SUCCESS)
        return FAIL; /*TODO: should it be UNAUTHORIZED?*/
    
    a = apr_array_make(temp_pool, 1, sizeof(char *));
    
    ret = ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK);

    if (ret != OK) {
        *retval = ret;
        return FAIL; /*above: module should return that error code*/
    }
    
    if (ap_should_client_block(r)) {
        char temp_buf[HUGE_STRING_LEN + 1];
        
        while ((res = ap_get_client_block(r, temp_buf, HUGE_STRING_LEN)) > 0) {
            *(temp_buf + res) = '\0';
            *(const char **)apr_array_push(a) =
                apr_pstrdup(temp_pool, temp_buf);
            /*this array is free'd before the function exits*/
        }
    } else {
#ifdef CSPACE_DEBUG
        printf("ERR: should_client_block false!\n");
#endif
        return FAIL;
    }

    *buf = apr_array_pstrcat(p, a, '\0'); /*use the request pool here*/
    
    apr_pool_destroy(temp_pool);

    return SUCC;
    /*return (res == OK);*/
}

static void set_header(char *key, char *val, void* table, void *pool)
{
	apr_table_t *t = (apr_table_t *)table;

    printf("Header %s:%s\n", key, val);

    apr_table_set(t, key, val);
    /*apr_pstrdup((apr_pool_t *)pool, key),
                  apr_pstrdup((apr_pool_t *)pool, val));*/
    /*apr_table_set(t, strdup(key), 
                     strdup(val));*/

    /*is dup necessary? update: NO! apr_table_set dups the key and the val*/
}

static int iterate_headers(void *rec, const char *key, const char *val)
{
    if (strstr(key, "cardspace"))
        return FAIL;
    else
        return SUCC;
}

/*with the use of subprocess_env, this function is useless now*/
static int check_valid_headers(request_rec *r)
{
    return apr_table_do(iterate_headers, NULL, r->subprocess_env, NULL);
}

static int set_cookie(request_rec *r, const char* session_id)
{
    /*apr_table_set(r->headers_out, SET_COOKIE, session_id);*/
    const char *cookie_header = NULL;
    char *cookies = NULL;
    size_t cookie_len = 0; /*header len*/

    cookie_header = apr_table_get(r->headers_out, HTTP_COOKIE_HEADER_OUT);

    if (cookie_header) {
        /*header already set*/
        if (strstr(cookie_header, SESSION_COOKIE)) {
            return SUCC; /*session_id is already set*/
        } else {
            /*session id not set, set it now*/
            cookie_len = strlen(cookie_header) - 
                         strlen(HTTP_COOKIE_HEADER_OUT);

            if (!(cookies = apr_palloc(r->pool, cookie_len + 3 +
                                                /*3 for ";=\0"*/
                                                strlen(SESSION_COOKIE) +
                                                MAX_SESSION_ID_LEN)))
                    return HTTP_INTERNAL_SERVER_ERROR;

            strncpy(cookies, cookie_header + strlen(HTTP_COOKIE_HEADER_OUT),
                    cookie_len);
            strncpy(cookies + cookie_len, ";", 1);
        }
    }
    else {
        if (!(cookies = apr_palloc(r->pool, cookie_len + 3 + /*3 for ";=\0"*/
                                            strlen(SESSION_COOKIE) +
                                            MAX_SESSION_ID_LEN)))
            return HTTP_INTERNAL_SERVER_ERROR;
    }
    /*no cookie header found, or cookie not in header*/
    
    strncpy(cookies + cookie_len, SESSION_COOKIE, strlen(SESSION_COOKIE));
    strncpy(cookies + cookie_len + strlen(SESSION_COOKIE), "=", 1);
    strncpy(cookies + cookie_len + 1 + strlen(SESSION_COOKIE), session_id,
            strlen(session_id));
    strncpy(cookies + cookie_len + 1 + strlen(SESSION_COOKIE) +
            strlen(session_id), "\0", 1);
    
    apr_table_unset(r->headers_out, HTTP_COOKIE_HEADER_OUT);
    apr_table_set(r->headers_out, HTTP_COOKIE_HEADER_OUT, cookies);

    return SUCC;
}

static int get_cookie(request_rec *r, char **cookies)
{
    const char *cookie_header = NULL;
    size_t cookie_len; /*header len*/

    cookie_header = apr_table_get(r->headers_in, HTTP_COOKIE_HEADER_IN);
    if (!cookie_header)
        return FAIL; /*no cookie header found*/
  
    cookie_len = strlen(cookie_header); /*- strlen(HTTP_COOKIE_HEADER_IN)*/
    if (!(*cookies = apr_palloc(r->pool, cookie_len)))
        return HTTP_INTERNAL_SERVER_ERROR;

    strncpy(*cookies, cookie_header /* + strlen(HTTP_COOKIE_HEADER_IN)*/,
            cookie_len);
    strncpy(*cookies + cookie_len, "\0", 1);
    
    return SUCC;
}

static char *get_session_id_from_cookies(char *cookies)
{
    char *ptr = NULL;
    if ((cookies) && (cookies = strstr(cookies, SESSION_COOKIE))) {
        /*cookie in hand*/
        if ((cookies) && (ptr = strchr(cookies, '='))) {
            /*cookie structure is fine, now get the value*/
            if ((ptr) && (ptr = strtok(ptr + 1, " ;\n\t\r\f"))) {
                /*ptr contains the valid session ID*/
                printf("session id received:%s\n", ptr);
                printf("all cookies received: %s\n", cookies);
                return ptr;
            }
        }
    }

    return NULL;
}

/*TODO:DONE make this to handle an excluded list instead of just one uri*/
static int excluded_uri(const char *uri, apr_array_header_t *excluded_list)
{
    if (find_uri(uri, excluded_list))
        return SUCC;
    else
        return FAIL;
}

static int cspace_login_arg_avail(request_rec *r)
{
    if (r->args) {
        if (strncmp(r->args, CSPACE_DEFAULT_LOGIN_ARG,
                    MYMIN(strlen(CSPACE_DEFAULT_LOGIN_ARG),
                          strlen(r->args))) == 0)
            return SUCC;
        else if (strstr(r->args, CSPACE_DEFAULT_LOGIN_AMP_ARG))
            return SUCC;
    }

    return FAIL;
}

static int handle_session_sso(request_rec *r, cspace_dir_cfg *dir_cfg,
                              cspace_svr_cfg *svr_cfg)
{
    /*TODO*/
    return UNAUTHORIZED_REDIRECT(r, NULL);
}

static int handle_nosession_sso(request_rec *r, cspace_dir_cfg *dir_cfg,
                                cspace_svr_cfg *svr_cfg)
{
    char *session_id = NULL;
    int valid_session = FAIL;
    int status = FAIL;
    char *cookies = NULL;
    void *session_ctx = NULL;

    status = get_cookie(r, &cookies);
    if (status == HTTP_INTERNAL_SERVER_ERROR)
        return status;

    session_ctx = session_ctx_create(r->pool, dir_cfg->session_file,
                                     svr_cfg->session_expire);

    /*NB: should we return HTTP_METHOD_NOT_ALLOWED if not POST?*/
    /* TODO: make the argument (tag identifying login page) a config option */
    if ((cookies) && (session_id = get_session_id_from_cookies(cookies))) {
        if ((valid_session = handle_session(session_id, SESSION_ID_LEN,
                                            session_ctx)))
            return OK;
    }

    if ((cspace_login_arg_avail(r)) && (r->method_number == M_POST)) {
       /*TODO*/ 
    }
    return UNAUTHORIZED_REDIRECT(r, NULL);
}

static int handle_session_nosso(request_rec *r, cspace_dir_cfg *dir_cfg,
                                cspace_svr_cfg *svr_cfg)
{
    char *session_id;
    int valid_session;
    int status;
    char *buf = NULL;
    int ret;
    const char *auth_state;
    char *cookies = NULL;
    void *session_ctx = NULL;

    if (excluded_uri(r->uri, dir_cfg->login_page))
        return OK;

    status = get_cookie(r, &cookies);
    if (status == HTTP_INTERNAL_SERVER_ERROR)
        return status;

    session_ctx = session_ctx_create(r->pool, dir_cfg->session_file,
                                     svr_cfg->session_expire);

    if ((cookies) && (session_id = get_session_id_from_cookies(cookies))) {
        if ((valid_session = handle_session(session_id, SESSION_ID_LEN,
                                            session_ctx)))
            return OK;
    }

    /*NB: should we return HTTP_METHOD_NOT_ALLOWED if not POST?*/
    /* TODO: make the argument (tag identifying login page) a config option */
    if ((cspace_login_arg_avail(r)) && (r->method_number == M_POST)) {
        if (cspace_read_request(r, &buf, &ret)) {
            /*TODO: xmlToken can be different! yet another config option*/
            if ((buf) && (strstr((const char *)buf, CSPACE_XML_TOKEN))) {
                int state = FAIL;
                process_ctx_t *ctx = NULL;

                cspace_decode_url((char *)buf);

                if (!(ctx = cspace_process_ctx_create((void *)(r->pool))))
                    return HTTP_INTERNAL_SERVER_ERROR;

                state = cspace_process_request(ctx, buf, svr_cfg->key_file,
                            dir_cfg->tvs, (void *)(r->subprocess_env),
                            &set_header);
                
                if (!state) {
                    /*TODO:clear all the headers added*/
                }

                auth_state = apr_table_get(r->subprocess_env,
                                           CSPACE_TVS_RES_STATE_HEADER);
                
                if ((auth_state) && 
                        (strcmp(auth_state, 
                                CSPACE_TVS_RES_STATE_SUCCESS) == 0)) {
                    if (!gen_session(SESSION_ID_LEN, session_ctx,
                                     &session_id))
                        return HTTP_INTERNAL_SERVER_ERROR;

                    if (session_id)
                        set_cookie(r, session_id);

                    return OK;
                } else {
                    return UNAUTHORIZED_REDIRECT(r, NULL);
                }
            }
        } else {
#ifdef CSPACE_DEBUG
            printf("ERR: request null! can't read?");
#endif
            if (ret != OK)
                return ret;
        }
    } 

    return UNAUTHORIZED_REDIRECT(r, NULL);
}

static int handle_nosession_nosso(request_rec *r, cspace_dir_cfg *dir_cfg,
                                  cspace_svr_cfg *svr_cfg)
{
    char *buf = NULL;
    int ret;

    /*NB: should we return HTTP_METHOD_NOT_ALLOWED if not POST?*/
    /* TODO: make the argument (tag identifying login page) a config option */
    if ((cspace_login_arg_avail(r)) && (r->method_number == M_POST)) {
        if (cspace_read_request(r, &buf, &ret)) {
            
            /*TODO: xmlToken can be different! yet another config option*/
            if ((buf) && (strstr((const char *)buf, CSPACE_XML_TOKEN))) {
                
                int state;
                process_ctx_t *ctx = NULL;

                cspace_decode_url((char *)buf);

                if (!(ctx = cspace_process_ctx_create((void *)(r->pool))))
                    return HTTP_INTERNAL_SERVER_ERROR;
                
                state = cspace_process_request(ctx, buf, svr_cfg->key_file,
                                               dir_cfg->tvs,
                                               (void *)(r->subprocess_env),
                                               &set_header);

                if (state)
                    return OK;
            }

        } else {
#ifdef CSPACE_DEBUG
            printf("ERR: request null! can't read?");
#endif
            
            /*Should it be UNAUTHORIZED_REDIRECT(r, NULL) that should go here*/
            if (ret != OK)
                return ret;
        }
    }

    return OK;
}

static int cspace_access_check(request_rec *r)
{
    cspace_dir_cfg *dir_cfg = NULL;
    cspace_svr_cfg *svr_cfg = NULL;

    dir_cfg = (cspace_dir_cfg *)ap_get_module_config(r->per_dir_config,
                                                     &cspace_module);
  
    svr_cfg = (cspace_svr_cfg *)ap_get_module_config(r->server->module_config,
                                                     &cspace_module);

#ifdef CSPACE_DEBUG
    printf("CSPACE_ACCESS_CHECK\n");
    cfg_svr_printf("\tsvr cfg:", svr_cfg, "\n");
    cfg_dir_printf("\tdir cfg:", dir_cfg, "\n");
    printf("\turi: %s\n", r->uri);
#endif
    
    if (!(dir_cfg->use_cspace_auth))
        return DECLINED;

    if (strcmp(ap_http_scheme(r), "https") != 0)
        return UNAUTHORIZED_REDIRECT(r, NULL);

    /* We no longer use the headers_in table,
     * hence this check is unnecessary
     */
    /*Don't send us CARDSPACE headers*/
    /*if (!check_valid_headers(r))
        return UNAUTHORIZED_REDIRECT(r, NULL); */

 
    if (dir_cfg->session) {
        if (dir_cfg->sso) {
            return handle_session_sso(r, dir_cfg, svr_cfg);
        } else {
            return handle_session_nosso(r, dir_cfg, svr_cfg);
        }
    } else {
    /*do not perform session_handling*/
        if (dir_cfg->sso) {
            return handle_nosession_sso(r, dir_cfg, svr_cfg);
        } else {
            return handle_nosession_nosso(r, dir_cfg, svr_cfg);
        }
    }
    
    return UNAUTHORIZED_REDIRECT(r, NULL);
}

static void cspace_reg_hooks(apr_pool_t *p)
{
    /*ap_hook_post_config(cspace_post_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_child_init(cspace_child_init, NULL, NULL, APR_HOOK_MIDDLE);*/
    ap_hook_access_checker(cspace_access_check, NULL, NULL, APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA cspace_module = {
    STANDARD20_MODULE_STUFF,
    cspace_dir_cfg_create,
    cspace_dir_cfg_merge,
    cspace_svr_cfg_create,
    cspace_svr_cfg_merge,
    cspace_cmds,
    cspace_reg_hooks
};

