/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.solutions.identity.relyingparty;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.security.WSConstants;
import org.apache.ws.security.processor.EncryptedKeyProcessor;
import org.apache.ws.security.util.WSSecurityUtil;
import org.apache.xml.security.Init;
import org.apache.xml.security.encryption.XMLCipher;
import org.apache.xml.security.keys.KeyInfo;
import org.apache.xml.security.signature.XMLSignature;
import org.apache.xml.security.utils.EncryptionConstants;
import org.opensaml.InvalidCryptoException;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLAttribute;
import org.opensaml.SAMLAttributeStatement;
import org.opensaml.SAMLException;
import org.opensaml.SAMLStatement;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.wso2.solutions.identity.i18n.Messages;

import javax.crypto.SecretKey;
import javax.xml.parsers.DocumentBuilderFactory;

import java.io.ByteArrayInputStream;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

/**
 * A SAML token is sent to a web application in a CardSpace login attempt
 * and this can be used for decryption and verification of those tokens. 
 */
public class TokenVerifier {

    private static Log log = LogFactory.getLog(TokenVerifier.class);

    private static Messages messages = Messages
            .getInstance(TokenVerifierConstants.RESOURCES);

    private Hashtable attributeTable = new Hashtable();

    private List certificates = new ArrayList();

    private Element keyInfoElement = null;
    
    private String issuerName = null;
    
    static {
        Init.init();
    }

    /**
     * Decrypt the given token (as a <code>java.lang.String</code> with the 
     * given private key.
     * @param token Serialized SAML token
     * @param serviceKey Private key to be used for decryption.
     * @return Decrypted SAML token element.
     * @throws RelyingPartyException
     */
    public Element decryptToken(String token, PrivateKey serviceKey)
            throws RelyingPartyException {

        try {

            if (log.isDebugEnabled()) {
                log.debug(messages.getMessage("receivedEncryuptedToken",
                        new String[] { token }));
            }

            ByteArrayInputStream bais = new ByteArrayInputStream(token
                    .getBytes());
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            Document doc = dbf.newDocumentBuilder().parse(bais);
            Element encryptedTokenElem = doc.getDocumentElement();

            // Decrypt element
            return this.decryptElement(serviceKey, encryptedTokenElem);

        } catch (Exception e) {
            throw new RelyingPartyException("verificationFailure", e);
        }

    }

    /**
     * Verify the signature of a given SAML token (decrypted)
     * @param decryptedElem SAML token element
     * @return true if verification is successful and false if unsuccessful.
     * @throws SAMLException
     */
    public boolean verifyDecryptedToken(Element decryptedElem)
            throws RelyingPartyException {
        try {
            if (log.isDebugEnabled()) {
                log.debug(messages.getMessage("verifyingDecryptedToken"));
            }
            
            SAMLAssertion assertion = new SAMLAssertion(decryptedElem);
            
            if (assertion.isSigned()) {
                assertion.verify();
                try {
                    Iterator ite = assertion.getX509Certificates();
                    while (ite.hasNext()) {
                        X509Certificate cert = (X509Certificate) ite.next();
                        this.certificates.add(cert);
                    }
                } catch (InvalidCryptoException e) {
                    // try to get the keyInfo
                    XMLSignature sigObject = (XMLSignature) assertion
                            .getNativeSignature();
                    if (sigObject != null) {
                        KeyInfo keyinfo = sigObject.getKeyInfo();
                        keyInfoElement = keyinfo.getElement();
                    } else {
                        throw e;
                    }
                }
            }

            Iterator statements = assertion.getStatements();
            while (statements.hasNext()) {
                SAMLStatement stmt = (SAMLStatement) statements.next();
                if (stmt instanceof SAMLAttributeStatement) {
                    Iterator attrs = ((SAMLAttributeStatement) stmt)
                            .getAttributes();
                    while (attrs.hasNext()) {
                        SAMLAttribute attr = (SAMLAttribute) attrs.next();
                        Object value = attr.getValues().next();
                        String name = attr.getName();

                        if (log.isDebugEnabled()) {
                            log.debug(messages.getMessage("samlAttrFound",
                                    new Object[] { name, value }));
                        }
                        this.attributeTable.put(name, value);
                    }
                }
            }
            
            issuerName = assertion.getIssuer();
            if(issuerName== null){
                throw new RelyingPartyException("issuerIsNull");
            }

        } catch (SAMLException e) {
            throw new RelyingPartyException("verificationFailure", e);
        }

        if (log.isDebugEnabled()) {
            log.debug(messages.getMessage("verifyingDecryptedTokenDone"));
        }

        // If we reach this point ... everything is fine :D
        return true;
    }

    private Element decryptElement(PrivateKey privKey, Element encryptedToken)
            throws Exception {

        if (log.isDebugEnabled()) {
            log.debug(messages.getMessage("decryptingToken"));
        }

        Element kiElem = (Element) encryptedToken.getElementsByTagNameNS(
                WSConstants.SIG_NS, "KeyInfo").item(0);
        Element encrKeyElem = (Element) kiElem.getElementsByTagNameNS(
                WSConstants.ENC_NS, EncryptionConstants._TAG_ENCRYPTEDKEY)
                .item(0);

        EncryptedKeyProcessor encrKeyProcessor = new EncryptedKeyProcessor();
        encrKeyProcessor.handleEncryptedKey(encrKeyElem, privKey);

        SecretKey secretKey = WSSecurityUtil.prepareSecretKey(
                EncryptionConstants.ALGO_ID_BLOCKCIPHER_AES128,
                encrKeyProcessor.getDecryptedBytes());

        XMLCipher cipher = XMLCipher.getInstance();
        cipher.init(XMLCipher.DECRYPT_MODE, secretKey);

        Document doc = cipher.doFinal(encryptedToken.getOwnerDocument(),
                encryptedToken);
        if (log.isDebugEnabled()) {
            log.debug(messages.getMessage("decryptingTokenDone"));
        }

        return doc.getDocumentElement();
    }

    /**
     * Returns the list of attributes extracted from the SAMLAttributeStatements
     * in the verified SAML assertion.
     * @return List of attributes as a  <code>java.util.Hashtable</code>
     */
    public Hashtable getAttributeTable() {
        return attributeTable;
    }

    public List getCertificates() {
        return certificates;
    }

    public Element getKeyInfoElement() {
        return keyInfoElement;
    }

    public String getIssuerName() {
        return issuerName;
    }

    public void setIssuerName(String issuer) {
        this.issuerName = issuer;
    }

}